from pathlib import Path
from typing import TextIO
from typing import cast

import logging

from logging import Logger
from logging import getLogger
from logging import config

from pkg_resources import resource_filename

from json import load as jsonLoad

from os import sep as osSep

from PyTravisCI import TravisCI
from PyTravisCI import defaults

from click import Context
from click import command
from click import option
from click import version_option
from click import get_current_context
from click import secho
from click import open_file
from click import style
from click import INT
from click import Path as clickPath
from click import clear as clickClear
from click import echo as clickEcho

from PyTravisCI.resource_types.builds import Builds
from PyTravisCI.resource_types.repository import Repository

from travisci.Preferences import Preferences
from travisci.SemanticVersion import SemanticVersion
from travisci.exceptions.UnsupportedOperation import UnsupportedOperation


class TravisCli:

    JSON_LOGGING_CONFIG_FILENAME: str = "loggingConfiguration.json"
    MADE_UP_PRETTY_MAIN_NAME:     str = "TravisCli"

    RESOURCES_PACKAGE_NAME: str = 'travisci.resources'
    RESOURCES_PATH:         str = f'travisci{osSep}resources'
    RESOURCE_ENV_VAR:       str = 'RESOURCEPATH'

    def __init__(self):

        self._setupSystemLogging()
        self.logger: Logger = getLogger(TravisCli.MADE_UP_PRETTY_MAIN_NAME)

        Preferences.determinePreferencesLocation()
        self._preferences: Preferences = Preferences()

        self._buildCount:   int    = 1
        self._repoSlugName: str    = ''
        self._versionFile:  Path = cast(Path, None)
        self._majorVersion = ''
        self._minorVersion = ''
        self._patchVersion = ''

    def runCommand(self):

        repoBuilds:      Builds          = self._getTravisBuilds()
        semanticVersion: SemanticVersion = self.__getCurrentVersion()

        semanticVersion = self._updateVersionNumber(semanticVersion=semanticVersion)

        self._updateBuildNumber(semanticVersion=semanticVersion, repoBuilds=repoBuilds)

    @property
    def buildCount(self) -> int:
        raise UnsupportedOperation('CLI properties are write-only')

    @buildCount.setter
    def buildCount(self, newValue: int):
        self._buildCount = newValue

    @property
    def repoSlugName(self) -> str:
        raise UnsupportedOperation('CLI properties are write-only')

    @repoSlugName.setter
    def repoSlugName(self, newValue: str):
        self._repoSlugName = newValue

    @property
    def versionFile(self) -> Path:
        raise UnsupportedOperation('CLI properties are write-only')

    @versionFile.setter
    def versionFile(self, file: Path):
        self._versionFile = file

    @property
    def majorVersion(self) -> str:
        raise UnsupportedOperation('CLI properties are write-only')

    @majorVersion.setter
    def majorVersion(self, newVersion: str):
        self._majorVersion = newVersion

    @property
    def minorVersion(self) -> str:
        raise UnsupportedOperation('CLI properties are write-only')

    @minorVersion.setter
    def minorVersion(self, newVersion: str):
        self._minorVersion = newVersion

    @property
    def patchVersion(self) -> str:
        raise UnsupportedOperation('CLI properties are write-only')

    @patchVersion.setter
    def patchVersion(self, newVersion: str):
        self._patchVersion = newVersion

    def _getTravisBuilds(self) -> Builds:
        """
        Get a set of builds from Travis CI for the selected repository

        Returns:  The Travis builds
        """
        travisciApiToken: str = self._preferences.travisciApiToken
        self.logger.debug(f'Running Command with token: {travisciApiToken}')

        travisCI: TravisCI = TravisCI(access_token=travisciApiToken, access_point=defaults.access_points.PRIVATE)

        params = {'limit': self._buildCount}
        repository: Repository = travisCI.get_repository(self._repoSlugName)
        travisBuilds: Builds = repository.get_builds(params=params)

        return travisBuilds

    def _updateVersionNumber(self, semanticVersion: SemanticVersion) -> SemanticVersion:
        """
        Only one of the 3 numbers is not None
        If the minor version is updated then patch version goes to zero
        if the major version is updated then both the minor and patch version go to zero

        Args:
            semanticVersion:  The version the update

        Returns:

        """

        if self._patchVersion is not None:
            semanticVersion.patch = self._patchVersion
        elif self._minorVersion is not None:
            semanticVersion.minor = self._minorVersion
            semanticVersion.patch = 0
        elif self._majorVersion is not None:
            semanticVersion.major = self._majorVersion
            semanticVersion.minor = 0
            semanticVersion.patch = 0

        return semanticVersion

    def _updateBuildNumber(self, semanticVersion: SemanticVersion, repoBuilds: Builds):

        highestBuildNumber: str = self.__getHighestBuildNumber(repoBuilds)

        highestBuildNumber = f'+.{highestBuildNumber}'      # Normalize it

        semanticVersion = self.__updateVersionFile(highestBuildNumber, semanticVersion)

        return semanticVersion

    def _setupSystemLogging(self):

        configFilePath: str = self._retrieveResourcePath(TravisCli.JSON_LOGGING_CONFIG_FILENAME)

        with open(configFilePath, 'r') as loggingConfigurationFile:
            configurationDictionary = jsonLoad(loggingConfigurationFile)

        config.dictConfig(configurationDictionary)
        logging.logProcesses = False
        logging.logThreads   = False

    def _retrieveResourcePath(self, bareFileName: str) -> str:

        # Use this method in Python 3.9
        # from importlib_resources import files
        # configFilePath: str  = files('travisci.resources').joinpath(TravisCli.JSON_LOGGING_CONFIG_FILENAME)

        try:
            fqFileName: str = resource_filename(TravisCli.RESOURCES_PACKAGE_NAME, bareFileName)
        except (ValueError, Exception):
            #
            # Maybe we are in an app
            #
            from os import environ
            pathToResources: str = environ.get(f'{TravisCli.RESOURCE_ENV_VAR}')
            fqFileName:      str = f'{pathToResources}{osSep}{TravisCli.RESOURCES_PATH}{osSep}{bareFileName}'

        return fqFileName

    def __getHighestBuildNumber(self, repoBuilds: Builds) -> str:
        """
        Searches the input list of repository builds and determines the largest build number

        Args:
            repoBuilds: The Travis CI repository builds

        Returns:  The string version of the build number
        """
        highestBuildNumber: str = '0'

        for build in repoBuilds:

            if int(build.number) > int(highestBuildNumber):
                highestBuildNumber = build.number

        self.logger.info(f'{highestBuildNumber=}')

        return highestBuildNumber

    def __getCurrentVersion(self) -> SemanticVersion:
        """
        Reads the version text file that is in semantic version format

        Returns:  The semantic version object that represents the current version stored in the text file
        """
        readDescriptor:  TextIO          = open_file(self._versionFile, mode='r')
        semanticVersion: SemanticVersion = SemanticVersion(readDescriptor.read())
        readDescriptor.close()

        secho(f'Old Version: {semanticVersion}')

        return semanticVersion

    def __updateVersionFile(self, normalizedBuildNumber: str, semanticVersion: SemanticVersion) -> SemanticVersion:
        """
        Updates the version text file

        Args:
            normalizedBuildNumber:  String set up as 'proper' build number
            semanticVersion:        The semantic version from the old text file
        """
        semanticBuildNbr      = semanticVersion.toBuildNumber(normalizedBuildNumber)
        semanticVersion.build = semanticBuildNbr
        secho(f'New Version: {semanticVersion}')

        writeDescriptor: TextIO = open_file(self._versionFile, mode='w')
        writeDescriptor.write(semanticVersion.__str__())
        writeDescriptor.close()

        return semanticVersion


@command()
@option('-b', '--build-count',     default=5,      type=INT, help='Number builds to check.')
@option('-r', '--repo-slug',   required=True,  help='something thing like hasii2011/PyUt.')
@option('-f', '--file',        default='travisci/resources/version.txt', type=clickPath(exists=True),  help='Relative location of version text file')
@option('--major-version',     required=False, type=INT, help='Change the major number to the specified one')
@option('--minor-version',     required=False, type=INT, help='Change the minor number to the specified one')
@option('--patch-version',     required=False, type=INT, help='Change the patch number to the specified one')
@version_option(version='0.3.2', message='%(version)s')
def commandHandler(build_count: int, repo_slug: str, file: TextIO, major_version: int, minor_version: int, patch_version: int):
    """
    Use this command to get the Travis CI build number of your project.  Assumes you are using Semantic Versioning
    """
    clickClear()
    clickEcho(style(f"Starting {TravisCli.MADE_UP_PRETTY_MAIN_NAME}", reverse=True))

    travisCmd: TravisCli = TravisCli()

    travisCmd.buildCount   = build_count
    travisCmd.repoSlugName = repo_slug
    travisCmd.versionFile  = file

    ctx: Context = get_current_context()
    if (major_version and minor_version) or (major_version and patch_version) or (minor_version and patch_version):
        clickEcho('You can only specify one of --major-version, --minor-version, or --patch-version')
        ctx.exit(1)

    travisCmd.majorVersion = major_version
    travisCmd.minorVersion = minor_version
    travisCmd.patchVersion = patch_version

    # Launch travisCmd
    travisCmd.runCommand()


if __name__ == "__main__":

    commandHandler()



from types import MethodType


class Singleton(object):
    """
    Base class for singleton classes.

    Any class derived from this one is a singleton. You can call its
    constructor multiple times, you will get only one instance.

    Note that `__init__` must not be defined. Use `init` instead.
    This is because `__init__` will always be called, thus reinitializing the
    state of your singleton each time you try to instantiate it.
    This implementation will only call `init` a single time

    To ensure that the `__init__` method is  defined,
    the singleton checks for its existence at first instantiation and raises an
    `AssertionError` if it finds it.

    Example::

        # This class is OK
        class NoInitClass(Singleton):
            def init(self, val):
                self.val = val

        # This class will raise AssertionError at first instantiation, because
        # of the __init__ method.
        class HasInitClass(Singleton):
            def __init__(self, val):
                self.val = val
    """
    def __new__(cls, *args, **kwds):
        """
        New operator for this base singleton class.
        Will return the singleton instance or create it if needed.
        """
        instance = cls.__dict__.get("__instance__")
        if instance is None:
            instance = object.__new__(cls)
            assert type(instance.__init__) != MethodType, f"Error, your singleton class {cls} cannot contain an '__init__' method."
            try:
                instance.init(*args, **kwds)
            except Exception as e:
                raise e
            cls.__instance__ = instance
        return instance

    def init(self, *args, **kwds):
        """
        Constructor for a singleton class
        """
        pass




from typing import List
from typing import Sequence
from typing import Union
from typing import cast

from logging import Logger
from logging import getLogger

from re import Match as regexMatch
from re import compile as regexCompile

from itertools import zip_longest as iZipLongest


class SemanticVersionError(Exception):
    pass


class SemanticVersion:
    """
    TODO:  This needs to be rationalized with the one in Pyut

    Inspired by https://github.com/keleshev/version

    A normal version number MUST take the form X.Y.Z where
    X, Y, and Z are non-negative integers. X is the major
    version, Y is the minor version, and Z is the patch
    version. Each element MUST increase numerically by
    increments of one. For instance: 6.9.0 -> 6.10.0 ->
    6z.11.0.

    A pre-release version MAY be denoted by appending a dash
    and a series of dot separated identifiers immediately
    following the patch version. Identifiers MUST be
    comprised of only ASCII alphanumerics and dash
    [0-9A-Za-z-].

    A build version MAY be denoted by appending a plus sign
    and a series of dot separated identifiers immediately
    following the patch version or pre-release version.
    Identifiers MUST be comprised of only ASCII
    alphanumerics and dash [0-9A-Za-z-].

    When a major version number is incremented, the minor
    version and patch version MUST be reset to zero. When a
    minor version number is incremented, the patch version
    MUST be reset to zero. For instance: 6.1.3 -> 7.0.0 and
    8.1.7 -> 8.2.0.

    """
    def __init__(self, version: str):

        self.logger: Logger = getLogger(__name__)

        self._re = regexCompile('^'
                                '(\d+)\.(\d+)\.(\d+)'     # minor, major, patch
                                '(-[0-9A-Za-z-\.]+)?'     # pre-release
                                '(\+[0-9A-Za-z-\.]+)?'    # build
                                '$')

        match: regexMatch = self._re.match(version)
        if match is None:
            raise SemanticVersionError(f'Invalid Version: `{version}`')

        mGroups: Sequence[str] = match.groups()

        self.major: int = int(mGroups[0])
        self.minor: int = int(mGroups[1])
        self.patch: int = int(mGroups[2])

        self.preRelease: List[Union[int, str]] = self._makeGroup(match.group(4))
        self.build:      List[Union[int, str]] = self._makeGroup(match.group(5))

    def toBuildNumber(self, strValue: str) -> List[Union[int, str]]:
        """
        Must be a number
        Args:
            strValue: The string

        Returns:  An opaque type that can be used to update the build number

        """
        return self._makeGroup(strValue)

    def _majorMinorPatch(self) -> List[int]:
        return [self.major, self.minor, self.patch]

    def _makeGroup(self, g: str) -> List[Union[int, str]]:
        if g is None:
            return []
        else:
            return list(map(SemanticVersion.safeInt, g[1:].split('.')))

    @staticmethod
    def safeInt(s):
        assert type(s) is str
        try:
            return int(s)
        except ValueError:
            return s

    def __eq__(self, other: object):

        if self._comparable(other) is False:
            return False
        else:
            other: SemanticVersion = cast(SemanticVersion, other)
            return all([self._majorMinorPatch() == other._majorMinorPatch(),
                        self.build == other.build,
                        self.preRelease == other.preRelease])

    def __lt__(self, other):

        if self._comparable(other) is False:
            return False

        if self._majorMinorPatch() == other._majorMinorPatch():

            if self.preRelease == other.preRelease:
                if self.build == other.build:
                    return False
                elif self.build and other.build:
                    return self._isSequenceLess(self.build, other.build)
                elif self.build or other.build:
                    return bool(other.build)
                assert False, 'Should not be reachable'
            elif self.preRelease and other.preRelease:
                return self._isSequenceLess(self.preRelease, other.preRelease)
            elif self.preRelease or other.preRelease:
                return bool(self.preRelease)
            assert False, 'Should not be reachable'
        else:
            return self._majorMinorPatch() < other._majorMinorPatch()

    def __gt__(self, other):
        if not self < other and not self == other:
            return True
        else:
            return False

    def _isSequenceLess(self, left, right) -> bool:

        assert {int, str} >= set(map(type, right))

        for lStr, rStr in iZipLongest(left, right):
            assert not (lStr is None and rStr is None)

            if lStr is None or rStr is None:
                return bool(lStr is None)

            if type(lStr) is int and type(rStr) is int:
                if lStr < rStr:
                    return True
            elif type(lStr) is int or type(rStr) is int:
                return type(lStr) is int
            elif lStr != rStr:
                return lStr < rStr

    def _comparable(self, other: object) -> bool:

        if isinstance(other, SemanticVersion):
            return True
        else:
            return False

    def __str__(self):
        mmp: List[int] = self._majorMinorPatch()
        s:   str       = f'{str(mmp[0])}.{str(mmp[1])}.{str(mmp[2])}'

        if len(self.preRelease) > 0:
            s += '-%s' % '.'.join(str(s) for s in self.preRelease)
        if len(self.build) > 0:
            s += '+%s' % '.'.join(str(s) for s in self.build)
        return s

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.__str__())




from logging import Logger
from logging import getLogger

from os import getenv
from os import sep as osSep

from sys import platform

from configparser import ConfigParser

from travisci.Singleton import Singleton

from travisci.exceptions.PreferencesLocationNotSet import PreferencesLocationNotSet


class Preferences(Singleton):

    THE_GREAT_MAC_PLATFORM: str = 'darwin'
    PREFERENCES_FILE_NAME:  str = '.travisci-cli.ini'

    TRAVIS_CI_SECTION:      str = 'TRAVISCI'

    TRAVISCI_API_TOKEN_KEY:  str = 'travisci_api_token'

    preferencesFileLocationAndName: str = None

    def init(self):

        self.logger:  Logger = getLogger(__name__)

        self._config: ConfigParser = ConfigParser()
        self._loadConfiguration()

    @staticmethod
    def determinePreferencesLocation():
        """
        This method MUST (I repeat MUST) be called before
        attempting to instantiate the preferences Singleton
        """
        if platform == "linux2" or platform == "linux" or platform == Preferences.THE_GREAT_MAC_PLATFORM:
            Preferences.preferencesFileLocationAndName = getenv("HOME") + osSep + Preferences.PREFERENCES_FILE_NAME
        else:
            Preferences.preferencesFileLocationAndName = Preferences.PREFERENCES_FILE_NAME

    @staticmethod
    def getPreferencesLocation():
        if Preferences.preferencesFileLocationAndName is None:
            raise PreferencesLocationNotSet()
        else:
            return Preferences.preferencesFileLocationAndName

    @property
    def travisciApiToken(self) -> str:
        return self._config.get(Preferences.TRAVIS_CI_SECTION, Preferences.TRAVISCI_API_TOKEN_KEY)

    @travisciApiToken.setter
    def travisciApiToken(self, newValue: str):
        self._config.set(Preferences.TRAVIS_CI_SECTION, Preferences.TRAVISCI_API_TOKEN_KEY, newValue)
        self.__saveConfig()

    def _loadConfiguration(self):
        """
        Load preferences from configuration file
        """
        self.__ensureConfigurationFileExists()
        self._config.read(Preferences.getPreferencesLocation())

        self.__createSectionIfNecessary(Preferences.TRAVIS_CI_SECTION)

        self.__createNeededConfigurationKeys()

        self.__saveConfig()

    def __ensureConfigurationFileExists(self):

        # Make sure that the configuration file exists
        # noinspection PyUnusedLocal
        try:
            f = open(Preferences.getPreferencesLocation(), "r")
            f.close()
        except (ValueError, Exception) as e:
            try:
                f = open(Preferences.getPreferencesLocation(), "w")
                f.write("")
                f.close()
                self.logger.warning(f'Preferences file re-created')
            except (ValueError, Exception) as e:
                self.logger.error(f"Error: {e}")
                return

    def __createSectionIfNecessary(self, sectionName: str):

        hasSection: bool = self._config.has_section(sectionName)
        self.logger.debug(f'hasSection: {hasSection} - {sectionName}')
        if hasSection is False:
            self._config.add_section(sectionName)

    def __createNeededConfigurationKeys(self):

        if self._config.has_option(Preferences.TRAVIS_CI_SECTION, Preferences.TRAVISCI_API_TOKEN_KEY) is False:
            self._config.set(Preferences.TRAVIS_CI_SECTION, Preferences.TRAVISCI_API_TOKEN_KEY, 'PutYourTravisCIKeyHere')

    def __saveConfig(self):
        """
        Save configuration data to the configuration file
        """
        f = open(Preferences.getPreferencesLocation(), "w")
        self._config.write(f)
        f.close()



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="traviscli",
    version="0.3.2",
    author_email='Humberto.A.Sanchez.II@gmail.com',
    description='A build help tool',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/hasii2011/travisci-cli",
    packages=[
        'travisci',
        'travisci.exceptions',
        'travisci.resources'
    ],
    package_data={'travisci.resources': ['loggingConfiguration.json', 'loggingConfiguration.json']},
    include_package_data=True,
    install_requires=['click', 'PyTravisCI'],
    entry_points='''
        [console_scripts]
        traviscli=travisci.TravisCli:commandHandler
    ''',
)
